#  Copyright 2021 (c) Hin-Tak Leung <htl10@users.sourceforge.net>
#  See The FreeType Project LICENSE for license terms.
#
#  HNParsePage class, for extracting text and image positions
import struct

class HNParsePage(object):
    def __init__(self, data, old_style=False):
        self.data = data
        self.data_length = len(data)
        self.characters = []
        self.figures = []
        self.stats = {}
        self.offset = 0
        def Text(self, code):
            try:
                self.characters.append(bytes([self.data[self.offset+5],self.data[self.offset+4]]).decode("gbk"))
            except IndexError: # short data, nothing to do
                pass
            except UnicodeDecodeError:
                # HTL: When cut-and-paste on Linux, these transform to GB18030,
                # but I believe they are OCR artifacts. Where they occur,
                # 0xA38D 0xA38a (always together) are line-breaks, and 0xA389, 0xA3A0
                # are tabs and spaces.
                hash = {
                    0xA389 : "\t",
                    0xA38a : "\n",
                    0xA38D : "\r",
                    0xA3A0 : " ",
                    # # GB18030
                    #0xA389 : "",
                    #0xA38a : "",
                    #0xA38D : "",
                    #0xA3A0 : "",
                }
                code = self.data[self.offset+5] * 256 + self.data[self.offset+4]
                try:
                    #self.characters.append("<0x%04X>\n" % code)
                    self.characters.append(hash[code])
                except KeyError:
                    self.characters.append("<0x%04X>\n" % code)
            self.offset += 6

        def TextMulti(self, code):
            self.offset += 2
            if (code == 0x8001):
                self.characters.append("\n")
            while (1):
                try:
                    if (self.data[self.offset+1] == 0x80):
                        break
                except IndexError: # short data, nothing to do
                    return
                try:
                    self.characters.append(bytes([self.data[self.offset+3],self.data[self.offset+2]]).decode("gbk"))
                except UnicodeDecodeError:
                    self.characters.append("<0x%04X>\n" % (self.data[self.offset+3] * 256 + self.data[self.offset+2]))
                except IndexError: # short data, nothing to do
                    return
                self.offset += 4

        def Figure(self, code):
            try:
                self.data[self.offset+25]
            except IndexError: # short data, nothing to do
                return
            (ignore1, offset_x, offset_y, size_x, size_y, int2, int3, int4, int5)= struct.unpack("<HHHHHIIII", self.data[self.offset:self.offset+26])
            # in units of 1/2.473 pixels
            self.figures.append([offset_x, offset_y, size_x, size_y])
            self.offset += 26

        if (not old_style):
            dispatch = {
                0x8001 : Text,
                0x800A : Figure,
            }
        else:
            dispatch = {
                0x8001 : TextMulti,
                0x8070 : TextMulti,
                0x800A : Figure,
            }
        dispatch_keys = dispatch.keys()

        while (self.offset <= self.data_length - 2):
            (dispatch_code,) = struct.unpack("H", self.data[self.offset:self.offset+2])
            self.offset += 2
            if (dispatch_code in dispatch_keys):
                dispatch[dispatch_code](self, dispatch_code)
            else:
                self.offset +=2
                if (dispatch_code in self.stats.keys()):
                    self.stats[dispatch_code] +=1
                else:
                    self.stats[dispatch_code] = 1

    @property
    def texts(self):
        text = ''.join(self.characters)
        text.replace('\x00', '')
        text.replace('\r', '')
        return text



from os import getenv
from pathlib import Path
from platform import system
from subprocess import PIPE, run
from sys import exit, stderr


def build(setup_kwargs):
    """编译二进制依赖

    读取环境变量 `LIBJBIG2DEC` 选择依赖库：

    1. `LIBJBIG2DEC=0` 或未设置，则使用 poppler 库
    2. `LIBJBIG2DEC=1` 或任何非 `0` 值，使用 libjbig2dec 库

    note: 使用 pdm build <https://pdm.fming.dev/latest/pyproject/build/#custom-file-generation>
    """
    src = Path()
    dst = Path("build")
    dlls = src / "dlls"
    s_dep = src / "caj2pdf" / "dep"
    o_bin = dst / "caj2pdf" / "dep" / "bin"
    if not o_bin.exists():
        o_bin.mkdir(parents=True)

    # input
    decode_jbig2data_x_cc = (s_dep / "decode_jbig2data_x.cc").as_posix()
    decode_jbig2data_cc = (s_dep / "decode_jbig2data.cc").as_posix()
    jbigdec_cc = (s_dep / "jbigdec.cc").as_posix()
    JBigDecode_cc = (s_dep / "JBigDecode.cc").as_posix()

    # output
    libjbigdec_so = (o_bin / "libjbigdec.so").as_posix()
    libjbig2codec_so = (o_bin / "libjbig2codec.so").as_posix()

    if system() == "Linux":
        if getenv("LIBJBIG2DEC", "0") == "0":
            build_poppler(
                # output
                libjbigdec_so,
                libjbig2codec_so,
                # input
                jbigdec_cc,
                JBigDecode_cc,
                decode_jbig2data_cc,
            )
        else:
            build_jbig2dec(
                # output
                libjbigdec_so,
                libjbig2codec_so,
                # input
                jbigdec_cc,
                JBigDecode_cc,
                decode_jbig2data_x_cc,
            )
    elif system() == "Windows":
        s_libjbigdec_32dll = dlls / "libjbigdec-w32.dll"
        s_libjbig2codec_32dll = dlls / "libjbig2codec-w32.dll"
        s_libjbigdec_64dll = dlls / "libjbigdec-w64.dll"
        s_libjbig2codec_64dll = dlls / "libjbig2codec-w64.dll"
        t_libjbigdec_32dll = (o_bin / "libjbigdec-w32.dll")
        t_libjbig2codec_32dll = (o_bin / "libjbig2codec-w32.dll")
        t_libjbigdec_64dll = (o_bin / "libjbigdec-w64.dll")
        t_libjbig2codec_64dll = (o_bin / "libjbig2codec-w64.dll")
        sources = [
            s_libjbigdec_32dll,
            s_libjbig2codec_32dll,
            s_libjbigdec_64dll,
            s_libjbig2codec_64dll,
        ]
        targets = [
            t_libjbigdec_32dll,
            t_libjbig2codec_32dll,
            t_libjbigdec_64dll,
            t_libjbig2codec_64dll,
        ]
        for t, s in zip(targets, sources):
            if not t.exists():
                t.hardlink_to(s)
                print(f"BUILD: {t.as_posix()}")
            else:
                print(f"EXIST: {t.as_posix()}")


def build_poppler(
    # output
    libjbigdec_so,
    libjbig2codec_so,
    # input
    jbigdec_cc,
    JBigDecode_cc,
    decode_jbig2data_cc,
):
    dep_checker = run(["pkg-config", "--libs", "--cflags", "poppler"], stdout=PIPE)
    if dep_checker.returncode != 0:
        print("poppler 未安装，请通过系统包管理器安装", file=stderr)
        exit(-1)

    cmd_jbigdec = [
        "cc",
        "-Wall",
        "-fPIC",
        "-shared",
        "-o",
        libjbigdec_so,
        jbigdec_cc,
        JBigDecode_cc,
    ]
    run(cmd_jbigdec)
    print("BUILD: {}".format(libjbigdec_so))

    poppler_flags = dep_checker.stdout.decode("utf-8").strip().split(" ")
    cmd_jbig2codec = [
        "cc",
        "-Wall",
        *poppler_flags,
        "-fPIC",
        "-shared",
        "-o",
        libjbig2codec_so,
        decode_jbig2data_cc,
    ]
    run(cmd_jbig2codec)
    print("BUILD: {}".format(libjbig2codec_so))


def build_jbig2dec(
    # output
    libjbigdec_so,
    libjbig2codec_so,
    # input
    jbigdec_cc,
    JBigDecode_cc,
    decode_jbig2data_x_cc,
):
    dep_checker = run(["pkg-config", "--libs", "--cflags", "jbig2dec"], stdout=PIPE)
    if dep_checker.returncode != 0:
        print("jbig2dec 未安装，请通过系统包管理器安装", file=stderr)
        exit(-1)

    cmd_jbigdec = [
        "cc",
        "-Wall",
        "-fPIC",
        "-shared",
        "-o",
        libjbigdec_so,
        jbigdec_cc,
        JBigDecode_cc,
    ]
    run(cmd_jbigdec)
    print("BUILD: {}".format(libjbigdec_so))

    jbig2dec_flags = dep_checker.stdout.decode("utf-8").strip().split(" ")
    cmd_jbig2codec = [
        "cc",
        "-Wall",
        *jbig2dec_flags,
        "-fPIC",
        "-shared",
        "-o",
        libjbig2codec_so,
        decode_jbig2data_x_cc,
    ]
    run(cmd_jbig2codec)
    print("BUILD: {}".format(libjbig2codec_so))



#!/usr/bin/env python3

import argparse
import os
import platform

from .cajparser import CAJParser
from .install import install_context_windows
from .utils import add_outlines
from .version import __version__


def main():
    parser = argparse.ArgumentParser(prog="caj2pdf")
    parser.add_argument("--version", action="version", version=f"%(prog)s v{__version__}")
    subparsers = parser.add_subparsers(help="commands", dest="command")

    show_parser = subparsers.add_parser("show", help="Show the information of the CAJ file.")
    show_parser.add_argument("input", help="Path to the CAJ file.")

    convert_parser = subparsers.add_parser("convert", help="Convert the CAJ file to PDF file.")
    convert_parser.add_argument("input", help="Path to the CAJ file.")
    convert_parser.add_argument("-o", "--output", help="Output path to the PDF file.", required=False)

    outlines_parser = subparsers.add_parser("outlines", help="Extract outlines from the CAJ file and add it to PDF file.")
    outlines_parser.add_argument("input", help="Path to the CAJ file.")
    outlines_parser.add_argument("-o", "--output", help="Path to the PDF file.", required=True)

    parse_parser = subparsers.add_parser("parse", help="Parse CAJ file for debugging/development")
    parse_parser.add_argument("input", help="Path to the CAJ file.")

    text_extract_parser = subparsers.add_parser("text-extract", help="Parse CAJ file for debugging/development")
    text_extract_parser.add_argument("input", help="Path to the CAJ file.")

    install_parser = subparsers.add_parser("install", help="install some system features, may need admin permission.")
    install_parser.add_argument("--dry-run", help="not do actually, show the effect.", action="store_true")
    args = parser.parse_args()

    if args.command == "show":
        caj = CAJParser(args.input)
        if caj.format == "PDF" or caj.format == "KDH":
            print("File: {0}\nType: {1}\n".format(args.input, caj.format))
        else:
            print("File: {0}\nType: {1}\nPage count: {2}\nOutlines count: {3}\n".format(
                args.input,
                caj.format,
                caj.page_num,
                caj.toc_num
            ))

    if args.command == "convert":
        caj = CAJParser(args.input)
        if args.output is None:
            if args.input.endswith(".caj"):
                args.output = args.input.replace(".caj", ".pdf")
            elif (len(args.input) > 4 and (args.input[-4] == '.' or args.input[-3] == '.') and not args.input.endswith(".pdf")):
                args.output = os.path.splitext(args.input)[0] + ".pdf"
            else:
                args.output = args.input + ".pdf"
        caj.convert(args.output)

    if args.command == "outlines":
        caj = CAJParser(args.input)
        if caj.format == "PDF" or caj.format == "KDH":
            raise SystemExit("Unsupported file type: {0}.".format(caj.format))
        toc = caj.get_toc()
        add_outlines(toc, args.output, "tmp.pdf")
        os.replace("tmp.pdf", args.output)

    if args.command == "text-extract":
        caj = CAJParser(args.input)
        caj.text_extract()

    if args.command == "parse":
        caj = CAJParser(args.input)
        caj.parse()

    if args.command == "install":
        if platform.system() == "Windows":
            install_context_windows(args.dry_run)
        else:
            raise NotImplementedError("Only support Windows now.")



"""提供给上下文菜单调用

目前仅支持 Windows
"""
import pathlib
import sys
import threading

from .cajparser import CAJParser
from .exceptions import Caj2PdfException


def main():
    try:
        app()
    except Exception:
        pass


def app():
    try:
        cajfilepath_str = sys.argv[1]
    except IndexError:
        raise Caj2PdfException(f"找不到caj文件，输入参数为：{sys.argv!r}")

    cajfilepath = pathlib.Path(cajfilepath_str)
    if not cajfilepath.exists():
        raise Caj2PdfException(f"caj文件不存在：{cajfilepath.as_posix()}")

    inputfile = str(cajfilepath)
    outputfile = f"{inputfile}.pdf"
    task = threading.Thread(
        group=None, target=convert_caj, args=(inputfile, outputfile)
    )
    alive = AliveStatus()
    task.start()
    alive.start()
    task.join()
    alive.finish()
    alive.join()


def convert_caj(inputfile, outputfile):
    caj = CAJParser(inputfile)
    caj.convert(outputfile)


class AliveStatus(threading.Thread):
    def __init__(self):
        super().__init__(group=None)
        self.finished = False

    def run(self):
        from time import sleep

        status = ("正在转换 <<<<<<", "正在转换 >>>>>>")
        i = 0b0
        while not self.finished:
            print(status[i & 0b1], end="\r", file=sys.stderr)
            i ^= 0b1
            sleep(0.2)

    def finish(self):
        self.finished = True



#!/usr/bin/env python3

#  Copyright 2020-2021 (c) Hin-Tak Leung <htl10@users.sourceforge.net>
#  See The FreeType Project LICENSE for license terms.
#
#  python ctypes module and short program decodes the image data in a CAJ file.

#  To build, copy "libreaderex_x64.so" from the Ubuntu AppImage
#  to the current directory.
#  (See "Analysing libreaderex" in the Wiki on how to)
#
#  Then, run
#
#       cc -Wall -fPIC --shared -o libjbigdec.so jbigdec.cc JBigDecode.cc

import importlib.resources
import os
import platform
import struct
from ctypes import *

arch = platform.architecture()
if (arch[1] == 'WindowsPE'):
    if (arch[0] == '64bit'):
        with importlib.resources.files(__package__) as pkg_dir:
            dllpath = (pkg_dir / "bin/libjbigdec-w64.dll").as_posix()
            libjbigdec = cdll.LoadLibrary(dllpath)
    else:
        with importlib.resources.files(__package__) as pkg_dir:
            dllpath = (pkg_dir / "bin/libjbigdec-w32.dll").as_posix()
            libjbigdec = cdll.LoadLibrary(dllpath)
else:
    with importlib.resources.files(__package__) as pkg_dir:
        dllpath = (pkg_dir / "bin/libjbigdec.so").as_posix()
        libjbigdec = cdll.LoadLibrary(dllpath)

#SaveJbigAsBmp = libjbigdec.SaveJbigAsBmp
#SaveJbigAsBmp.restype = None
#SaveJbigAsBmp.argtypes = [c_void_p, c_int, c_char_p]

#SaveJbig2AsBmp = libjbigdec.SaveJbig2AsBmp
#SaveJbig2AsBmp.restype = None
#SaveJbig2AsBmp.argtypes = [c_void_p, c_int, c_char_p]

jbigDecode = libjbigdec.jbigDecode
jbigDecode.restype = None
jbigDecode.argtypes = [c_void_p, c_int, c_int, c_int, c_int, c_void_p]

class CImage:
    def __init__(self, buffer):
        self.buffer = buffer
        self.buffer_size=len(buffer)
        (self.width, self.height,
         self.num_planes, self.bits_per_pixel) = struct.unpack("<IIHH", buffer[4:16])
        self.bytes_per_line = ((self.width * self.bits_per_pixel + 31) >> 5) << 2

    def DecodeJbig(self):
        out = create_string_buffer(self.height * self.bytes_per_line)
        jbigDecode(self.buffer[48:], self.buffer_size-48, self.height, self.width, self.bytes_per_line, out)
        return out

if __name__ == '__main__':
    import os
    import sys

    if len(sys.argv) < 3:
        print("Usage: %s input output" % sys.argv[0])
        sys.exit()

    f = open(sys.argv[1], "rb")
    buffer_size = os.stat(sys.argv[1]).st_size
    buffer = f.read()

    #SaveJbigAsBmp(buffer, buffer_size, sys.argv[2].encode("ascii"))

    cimage = CImage(buffer)
    out = cimage.DecodeJbig()

    # PBM is only padded to 8 rather than 32.
    # If the padding is larger, write padded file.
    width = cimage.width
    if (cimage.bytes_per_line > ((cimage.width +7) >> 3)):
        width = cimage.bytes_per_line << 3

    fout = open(sys.argv[2].replace(".bmp", ".pbm"), "wb")
    fout.write("P4\n".encode("ascii"))
    fout.write(("%d %d\n" % (width, cimage.height)).encode("ascii"))
    fout.write(out)
    fout.close()



import os
import struct
from shutil import copy
from subprocess import STDOUT, CalledProcessError, check_output

from PyPDF2.errors import PdfReadError

from .utils import (
    add_outlines,
    fnd,
    fnd_all,
    fnd_rvrs,
    fnd_unuse_no,
    find_redundant_images,
)

KDH_PASSPHRASE = b"FZHMEI"

printables = "".join(
    [
        (len(repr(chr(x))) == 3) and (x != 47) and (x < 128) and chr(x) or "."
        for x in range(256)
    ]
)

image_type = {0: "JBIG", 1: "JPEG", 2: "JPEG", 3: "JBIG2"}  # up-side-down


class CAJParser(object):
    def __init__(self, filename):
        self.filename = filename
        try:
            with open(filename, "rb") as caj:
                caj_read4 = caj.read(4)
                if caj_read4[0:1] == b"\xc8":
                    self.format = "C8"
                    self._PAGE_NUMBER_OFFSET = 0x08
                    self._TOC_NUMBER_OFFSET = 0  # No TOC
                    self._TOC_END_OFFSET = 0x50
                    self._PAGEDATA_OFFSET = self._TOC_END_OFFSET + 20 * self.page_num
                    return
                if caj_read4[0:2] == b"HN":
                    if (
                        caj.read(2) == b"\xc8\x00"
                    ):  # Most of them are: 90 01, handled later
                        self.format = "HN"
                        self._PAGE_NUMBER_OFFSET = 0x90
                        self._TOC_NUMBER_OFFSET = 0
                        self._TOC_END_OFFSET = 0xD8
                        self._PAGEDATA_OFFSET = (
                            self._TOC_END_OFFSET + 20 * self.page_num
                        )
                        return
                fmt = (
                    struct.unpack("4s", caj_read4)[0]
                    .replace(b"\x00", b"")
                    .decode("gb18030")
                )
            if fmt == "CAJ":
                self.format = "CAJ"
                self._PAGE_NUMBER_OFFSET = 0x10
                self._TOC_NUMBER_OFFSET = 0x110
            elif fmt == "HN":
                self.format = "HN"
                self._PAGE_NUMBER_OFFSET = 0x90
                self._TOC_NUMBER_OFFSET = 0x158

                # TOC = [toc_num] followed by [toc_entry * toc_num]
                # followed by [Page Info struct (20-byte) * page_num], followed by Page Data
                self._TOC_END_OFFSET = (
                    self._TOC_NUMBER_OFFSET + 4 + 0x134 * self.toc_num
                )
                self._PAGEDATA_OFFSET = self._TOC_END_OFFSET + 20 * self.page_num
            elif fmt == "%PDF":
                self.format = "PDF"
            elif fmt == "KDH ":
                self.format = "KDH"
            elif fmt == "TEB":
                self.format = "TEB"
            else:
                self.format = None
                raise SystemExit("Unknown file type.")
        except UnicodeDecodeError:
            raise SystemExit("Unknown file type.")

    @property
    def page_num(self):
        with open(self.filename, "rb") as caj:
            caj.seek(self._PAGE_NUMBER_OFFSET)
            [page_num] = struct.unpack("i", caj.read(4))
            return page_num

    @property
    def toc_num(self):
        if self._TOC_NUMBER_OFFSET == 0:
            return 0
        with open(self.filename, "rb") as caj:
            caj.seek(self._TOC_NUMBER_OFFSET)
            [toc_num] = struct.unpack("i", caj.read(4))
            return toc_num

    def get_toc(self, verbose=False):
        toc = []
        if self._TOC_NUMBER_OFFSET == 0:
            return toc
        with open(self.filename, "rb") as caj:
            for i in range(self.toc_num):
                caj.seek(self._TOC_NUMBER_OFFSET + 4 + 0x134 * i)
                toc_bytes = struct.unpack("256s24s12s12si", caj.read(0x134))
                ttl_end = toc_bytes[0].find(b"\x00")
                title = toc_bytes[0][0:ttl_end].decode("gb18030").encode("utf-8")
                pg_end = toc_bytes[2].find(b"\x00")
                page = int(toc_bytes[2][0:pg_end])
                level = toc_bytes[4]
                toc_entry = {"title": title, "page": page, "level": level}
                if verbose:
                    print("   " * (level - 1), title.decode("utf-8"))
                toc.append(toc_entry)
            if verbose:
                print(
                    "TOC END: 0x%04X"
                    % (self._TOC_NUMBER_OFFSET + 4 + 0x134 * self.toc_num)
                )
        return toc

    def output_toc(self, dest):
        toc_items = self.get_toc()
        with open(dest, "wb") as f:
            for toc in toc_items:
                f.write(
                    b"    " * (toc["level"] - 1)
                    + toc["title"]
                    + b"    "
                    + str(toc["page"]).encode("utf-8")
                    + b"\n"
                )

    def convert(self, dest):
        if self.format == "CAJ":
            self._convert_caj(dest)
        elif self.format == "HN":
            self._convert_hn(dest)
        elif self.format == "C8":
            self._convert_hn(dest)
        elif self.format == "PDF":
            self._convert_pdf(dest)
        elif self.format == "KDH":
            self._convert_kdh(dest)

    def parse(self):
        if self.format == "CAJ":
            pass
        elif self.format == "HN":
            self._parse_hn()
        elif self.format == "C8":
            self._parse_hn()
        elif self.format == "PDF":
            pass
        elif self.format == "KDH":
            pass

    def text_extract(self):
        if self.format == "CAJ":
            pass
        if self.format == "HN":
            self._text_extract_hn()
        elif self.format == "C8":
            self._text_extract_hn()
        elif self.format == "PDF":
            pass
        elif self.format == "KDH":
            pass

    def _convert_caj(self, dest):
        caj = open(self.filename, "rb")

        # Extract original PDF data (and add header)
        caj.seek(self._PAGE_NUMBER_OFFSET + 4)
        [pdf_start_pointer] = struct.unpack("i", caj.read(4))
        caj.seek(pdf_start_pointer)
        [pdf_start] = struct.unpack("i", caj.read(4))
        pdf_end = fnd_all(caj, b"endobj")[-1] + 6
        pdf_length = pdf_end - pdf_start
        caj.seek(pdf_start)
        pdf_data = b"%PDF-1.3\r\n" + caj.read(pdf_length) + b"\r\n"
        with open("pdf.tmp", "wb") as f:
            f.write(pdf_data)
        pdf = open("pdf.tmp", "rb")

        # deal with disordered PDF data
        endobj_addr = fnd_all(pdf, b"endobj")
        obj_no = []
        for addr in endobj_addr:
            startobj = fnd_rvrs(pdf, b" 0 obj", addr)
            startobj1 = fnd_rvrs(pdf, b"\r", startobj)
            startobj2 = fnd_rvrs(pdf, b"\n", startobj)
            startobj = max(startobj1, startobj2)
            length = fnd(pdf, b" ", startobj) - startobj
            pdf.seek(startobj)
            [no] = struct.unpack(str(length) + "s", pdf.read(length))
            if int(no) not in obj_no:
                obj_no.append(int(no))
                obj_len = addr - startobj + 6
                pdf.seek(startobj)
                [obj] = struct.unpack(str(obj_len) + "s", pdf.read(obj_len))

        # Add Catalog (find obj_no of pages)
        inds_addr = [i + 8 for i in fnd_all(pdf, b"/Parent")]
        inds = []
        for addr in inds_addr:
            length = fnd(pdf, b" ", addr) - addr
            pdf.seek(addr)
            [ind] = struct.unpack(str(length) + "s", pdf.read(length))
            inds.append(int(ind))
        # get pages_obj_no list containing distinct elements
        # & find missing pages object(s) -- top pages object(s) in pages_obj_no
        pages_obj_no = []
        top_pages_obj_no = []
        for ind in inds:
            if (ind not in pages_obj_no) and (ind not in top_pages_obj_no):
                if fnd(pdf, bytes("\r{0} 0 obj".format(ind), "utf-8")) == -1:
                    top_pages_obj_no.append(ind)
                else:
                    pages_obj_no.append(ind)
        single_pages_obj_missed = len(top_pages_obj_no) == 1
        multi_pages_obj_missed = len(top_pages_obj_no) > 1
        # generate catalog object
        catalog_obj_no = fnd_unuse_no(obj_no, top_pages_obj_no)
        obj_no.append(catalog_obj_no)
        root_pages_obj_no = None
        if multi_pages_obj_missed:
            root_pages_obj_no = fnd_unuse_no(obj_no, top_pages_obj_no)
        elif single_pages_obj_missed:
            root_pages_obj_no = top_pages_obj_no[0]
            top_pages_obj_no = pages_obj_no
        else:  # root pages object exists, then find the root pages object #
            found = False
            for pon in pages_obj_no:
                tmp_addr = fnd(pdf, bytes("\r{0} 0 obj".format(pon), "utf-8"))
                while True:
                    pdf.seek(tmp_addr)
                    [_str] = struct.unpack("6s", pdf.read(6))
                    if _str == b"Parent":
                        break
                    elif _str == b"endobj":
                        root_pages_obj_no = pon
                        found = True
                        break
                    tmp_addr = tmp_addr + 1
                if found:
                    break
        catalog = bytes(
            "{0} 0 obj\r<</Type /Catalog\r/Pages {1} 0 R\r>>\rendobj\r".format(
                catalog_obj_no, root_pages_obj_no
            ),
            "utf-8",
        )
        pdf_data += catalog
        pdf.close()
        with open("pdf.tmp", "wb") as f:
            f.write(pdf_data)
        pdf = open("pdf.tmp", "rb")

        # Add Pages obj and EOF mark
        # if root pages object exist, pass
        # deal with single missing pages object
        if single_pages_obj_missed or multi_pages_obj_missed:
            inds_str = ["{0} 0 R".format(i) for i in top_pages_obj_no]
            kids_str = "[{0}]".format(" ".join(inds_str))
            pages_str = "{0} 0 obj\r<<\r/Type /Pages\r/Kids {1}\r/Count {2}\r>>\rendobj\r".format(
                root_pages_obj_no, kids_str, self.page_num
            )
            pdf_data += bytes(pages_str, "utf-8")
            pdf.close()
            with open("pdf.tmp", "wb") as f:
                f.write(pdf_data)
            pdf = open("pdf.tmp", "rb")
        # deal with multiple missing pages objects
        if multi_pages_obj_missed:
            kids_dict = {i: [] for i in top_pages_obj_no}
            count_dict = {i: 0 for i in top_pages_obj_no}
            for tpon in top_pages_obj_no:
                kids_addr = fnd_all(pdf, bytes("/Parent {0} 0 R".format(tpon), "utf-8"))
                for kid in kids_addr:
                    ind = fnd_rvrs(pdf, b"obj", kid) - 4
                    addr = fnd_rvrs(pdf, b"\r", ind)
                    length = fnd(pdf, b" ", addr) - addr
                    pdf.seek(addr)
                    [ind] = struct.unpack(str(length) + "s", pdf.read(length))
                    kids_dict[tpon].append(int(ind))
                    type_addr = fnd(pdf, b"/Type", addr) + 5
                    tmp_addr = fnd(pdf, b"/", type_addr) + 1
                    pdf.seek(tmp_addr)
                    [_type] = struct.unpack("5s", pdf.read(5))
                    if _type == b"Pages":
                        cnt_addr = fnd(pdf, b"/Count ", addr) + 7
                        pdf.seek(cnt_addr)
                        [_str] = struct.unpack("1s", pdf.read(1))
                        cnt_len = 0
                        while _str not in [b" ", b"\r", b"/"]:
                            cnt_len += 1
                            pdf.seek(cnt_addr + cnt_len)
                            [_str] = struct.unpack("1s", pdf.read(1))
                        pdf.seek(cnt_addr)
                        [cnt] = struct.unpack(str(cnt_len) + "s", pdf.read(cnt_len))
                        count_dict[tpon] += int(cnt)
                    else:  # _type == b"Page"
                        count_dict[tpon] += 1
                kids_no_str = ["{0} 0 R".format(i) for i in kids_dict[tpon]]
                kids_str = "[{0}]".format(" ".join(kids_no_str))
                pages_str = "{0} 0 obj\r<<\r/Type /Pages\r/Kids {1}\r/Count {2}\r>>\rendobj\r".format(
                    tpon, kids_str, count_dict[tpon]
                )
                pdf_data += bytes(pages_str, "utf-8")
        pdf_data += bytes("\n%%EOF\r", "utf-8")
        pdf.close()
        with open("pdf.tmp", "wb") as f:
            f.write(pdf_data)

        # Use mutool to repair xref
        try:
            check_output(["mutool", "clean", "pdf.tmp", "pdf_toc.pdf"], stderr=STDOUT)
        except CalledProcessError as e:
            print(e.output.decode("utf-8"))
            raise SystemExit(
                "Command mutool returned non-zero exit status " + str(e.returncode)
            )

        # Add Outlines
        try:
            add_outlines(self.get_toc(), "pdf_toc.pdf", dest)
        except PdfReadError as e:
            print("PdfReadError:", str(e))
            copy("pdf_toc.pdf", dest)
            pass
        os.remove("pdf.tmp")
        os.remove("pdf_toc.pdf")

    def _convert_hn(self, dest):
        caj = open(self.filename, "rb")
        image_list = []

        import zlib

        from .pdfwutils import Colorspace, ImageFormat, convert_ImageList

        for i in range(self.page_num):
            caj.seek(self._TOC_END_OFFSET + i * 20)
            [
                page_data_offset,
                size_of_text_section,
                images_per_page,
                page_no,
                unk2,
                next_page_data_offset,
            ] = struct.unpack("iihhii", caj.read(20))
            caj.seek(page_data_offset)
            text_header_read32 = caj.read(32)
            if (text_header_read32[8:20] == b"COMPRESSTEXT") or (
                text_header_read32[0:12] == b"COMPRESSTEXT"
            ):
                coff = 8
                if text_header_read32[0:12] == b"COMPRESSTEXT":
                    coff = 0
                [expanded_text_size] = struct.unpack(
                    "i", text_header_read32[12 + coff : 16 + coff]
                )
                caj.seek(page_data_offset + 16 + coff)
                data = caj.read(size_of_text_section - 16 - coff)
                output = zlib.decompress(data, bufsize=expanded_text_size)
                if len(output) != expanded_text_size:
                    raise SystemExit("Unexpected:", len(output), expanded_text_size)
            else:
                caj.seek(page_data_offset)
                output = caj.read(size_of_text_section)
            from .HNParsePage import HNParsePage

            page_style = next_page_data_offset > page_data_offset
            page_data = HNParsePage(output, page_style)

            current_offset = page_data_offset + size_of_text_section
            (found, images_per_page) = find_redundant_images(
                caj, current_offset, images_per_page
            )
            if found:
                print(
                    "Page %d, skipping %d redundant images"
                    % (i + 1, images_per_page * (images_per_page - 1))
                )

            if images_per_page > 1:
                if len(page_data.figures) == images_per_page:
                    if (page_data.figures[0][0] == 0) and (
                        page_data.figures[0][1] == 0
                    ):
                        image_list.append(None)
                        image_list.append(page_data.figures)
                    else:
                        print(
                            "Page %d, Image Count %d, first image not at origin, expanding to %d pages"
                            % (i + 1, len(page_data.figures), images_per_page)
                        )
                else:
                    print(
                        "Page %d, Image Count %d != %d"
                        % (i + 1, len(page_data.figures), images_per_page)
                    )
                    if len(page_data.figures) > images_per_page:
                        print("\tTruncating Page %d," % (i + 1), page_data.figures)
                        image_list.append(None)
                        image_list.append(page_data.figures[0:images_per_page])
                    else:
                        print(
                            "Page %d expanding to %d separate image pages"
                            % (i + 1, images_per_page)
                        )
            elif images_per_page == 1:
                if (len(page_data.figures) == 0) or (
                    (len(page_data.figures) > 0)
                    and (
                        not (
                            (page_data.figures[0][0] == 0)
                            and (page_data.figures[0][1] == 0)
                        )
                    )
                ):
                    print(
                        "Page %d possibly text-only + single figure(%d)"
                        % (i + 1, len(page_data.figures))
                    )
            else:
                # don't care about images_per_page == 0
                pass
            for j in range(images_per_page):
                caj.seek(current_offset)
                read32 = caj.read(32)
                [
                    image_type_enum,
                    offset_to_image_data,
                    size_of_image_data,
                ] = struct.unpack("iii", read32[0:12])
                if offset_to_image_data != current_offset + 12:
                    raise SystemExit("unusual image offset")
                caj.seek(offset_to_image_data)
                image_data = caj.read(size_of_image_data)
                current_offset = offset_to_image_data + size_of_image_data
                if image_type[image_type_enum] == "JBIG":
                    from .dep.jbigdec import CImage

                    cimage = CImage(image_data)
                    out = cimage.DecodeJbig()
                    # PBM is only padded to 8 rather than 32.
                    # If the padding is larger, write padded file.
                    width = cimage.width
                    if cimage.bytes_per_line > ((cimage.width + 7) >> 3):
                        width = cimage.bytes_per_line << 3
                    image_item = (
                        Colorspace.P,
                        (300, 300),
                        ImageFormat.PBM,
                        zlib.compress(out),
                        width,
                        cimage.height,
                        [0xFFFFFF, 0],
                        False,
                        1,
                        0,
                    )
                elif image_type[image_type_enum] == "JBIG2":
                    from .dep.jbig2dec import CImage

                    cimage = CImage(image_data)
                    out = cimage.DecodeJbig2()
                    # PBM is only padded to 8 rather than 32.
                    # If the padding is larger, write padded file.
                    width = cimage.width
                    if cimage.bytes_per_line > ((cimage.width + 7) >> 3):
                        width = cimage.bytes_per_line << 3
                    image_item = (
                        Colorspace.P,
                        (300, 300),
                        ImageFormat.PBM,
                        zlib.compress(out),
                        width,
                        cimage.height,
                        [0xFFFFFF, 0],
                        False,
                        1,
                        0,
                    )
                elif image_type[image_type_enum] == "JPEG":
                    colorspace = Colorspace.RGB
                    component = 3
                    # stock libjpeg location
                    (
                        SOFn,
                        frame_length,
                        bits_per_pixel,
                        height,
                        width,
                        component,
                    ) = struct.unpack(">HHBHHB", image_data[158:168])
                    if SOFn != 0xFFC0:
                        # "Intel(R) JPEG Library" location
                        (
                            SOFn,
                            frame_length,
                            bits_per_pixel,
                            height,
                            width,
                            component,
                        ) = struct.unpack(">HHBHHB", image_data[0x272:0x27C])
                        if SOFn != 0xFFC0:
                            # neither works, try brute-force
                            import imagesize
                            from PIL import Image as pilimage

                            with open(".tmp.jpg", "wb") as f:
                                f.write(image_data)
                                (width, height) = imagesize.get(".tmp.jpg")
                                pim = pilimage.open(".tmp.jpg")
                                if pim.mode == "L":
                                    component = 1
                            os.remove(".tmp.jpg")
                    if image_type_enum == 1:
                        # non-inverted JPEG Images
                        height = -height
                    if component == 1:
                        colorspace = Colorspace.L
                    image_item = (
                        colorspace,
                        (300, 300),
                        ImageFormat.JPEG,
                        image_data,
                        width,
                        height,
                        [],
                        False,
                        8,
                        0,
                    )
                else:
                    raise SystemExit("Unknown Image Type %d" % (image_type_enum))
                image_list.append(image_item)
        if len(image_list) == 0:
            raise SystemExit("File is pure-text HN; cannot convert to pdf")
        pdf_data = convert_ImageList(image_list)
        with open("pdf_toc.pdf", "wb") as f:
            f.write(pdf_data)
        # Add Outlines
        add_outlines(self.get_toc(), "pdf_toc.pdf", dest)
        os.remove("pdf_toc.pdf")

    def _text_extract_hn(self):
        if self._TOC_NUMBER_OFFSET > 0:
            self.get_toc(verbose=True)
        caj = open(self.filename, "rb")

        for i in range(self.page_num):
            caj.seek(self._TOC_END_OFFSET + i * 20)
            [
                page_data_offset,
                size_of_text_section,
                images_per_page,
                page_no,
                unk2,
                next_page_data_offset,
            ] = struct.unpack("iihhii", caj.read(20))
            caj.seek(page_data_offset)
            text_header_read32 = caj.read(32)
            if (text_header_read32[8:20] == b"COMPRESSTEXT") or (
                text_header_read32[0:12] == b"COMPRESSTEXT"
            ):
                coff = 8
                if text_header_read32[0:12] == b"COMPRESSTEXT":
                    coff = 0
                [expanded_text_size] = struct.unpack(
                    "i", text_header_read32[12 + coff : 16 + coff]
                )
                import zlib

                caj.seek(page_data_offset + 16 + coff)
                data = caj.read(size_of_text_section - 16 - coff)
                output = zlib.decompress(data, bufsize=expanded_text_size)
                if len(output) != expanded_text_size:
                    raise SystemExit("Unexpected:", len(output), expanded_text_size)
            else:
                caj.seek(page_data_offset)
                output = caj.read(size_of_text_section)
            from .HNParsePage import HNParsePage

            page_style = next_page_data_offset > page_data_offset
            page_data = HNParsePage(output, page_style)
            print("Text on Page %d:" % (i + 1))
            print(page_data.texts)
            # print("Figures:\n", page_data.figures)

    def _parse_hn(self):
        if self._TOC_NUMBER_OFFSET > 0:
            self.get_toc(verbose=True)
        caj = open(self.filename, "rb")

        for i in range(self.page_num):
            caj.seek(self._TOC_END_OFFSET + i * 20)
            print(
                "Reading Page Info struct #%d at offset 0x%04X"
                % (i + 1, self._TOC_END_OFFSET + i * 20)
            )
            [
                page_data_offset,
                size_of_text_section,
                images_per_page,
                page_no,
                unk2,
                next_page_data_offset,
            ] = struct.unpack("iihhii", caj.read(20))
            print(
                "unknown page struct members = (%d %d)" % (unk2, next_page_data_offset)
            )
            # All 71: 1,0,0
            print("Page Number %d Data offset = 0x%04X" % (page_no, page_data_offset))
            caj.seek(page_data_offset)
            text_header_read32 = caj.read(32)
            print("Page Text Header dump:\n", self.dump(text_header_read32), sep="")
            # The first 8 bytes are always: 03 80 XX 16 03 80 XX XX,
            # the last one 20 or 21, but the first two can be any.
            # 48/71 has: 03 80 E0 16 03 80 F7 20, the rest uniq
            if (text_header_read32[8:20] == b"COMPRESSTEXT") or (
                text_header_read32[0:12] == b"COMPRESSTEXT"
            ):
                coff = 8
                if text_header_read32[0:12] == b"COMPRESSTEXT":
                    coff = 0
                # expanded_text_size seems to be always about 2-3 times size_of_text_section, so this is a guess.
                [expanded_text_size] = struct.unpack(
                    "i", text_header_read32[12 + coff : 16 + coff]
                )
                import zlib

                caj.seek(page_data_offset + 16 + coff)
                data = caj.read(size_of_text_section - 16 - coff)
                output = zlib.decompress(data, bufsize=expanded_text_size)
                if len(output) != expanded_text_size:
                    print("Unexpected:", len(output), expanded_text_size)
                print(
                    "Page Text Header COMPRESSTEXT:\n",
                    self.dump(output, GB=True),
                    sep="",
                )
                for x in range(len(output) >> 4):
                    try:
                        print(
                            bytes([output[(x << 4) + 7], output[(x << 4) + 6]]).decode(
                                "gbk"
                            ),
                            end="",
                        )
                    except UnicodeDecodeError:
                        print(self.dump(output[x << 4 : (x + 1) << 4]))
                print()
            else:
                caj.seek(page_data_offset)
                output = caj.read(size_of_text_section)
                print(
                    "Page Text Header non-COMPRESSTEXT:\n",
                    self.dump(output, GB=True),
                    sep="",
                )
            from .HNParsePage import HNParsePage

            page_style = next_page_data_offset > page_data_offset
            page_data = HNParsePage(output, page_style)
            print("Text:\n", page_data.texts)
            print("Figures:\n", page_data.figures)
            current_offset = page_data_offset + size_of_text_section
            for j in range(images_per_page):
                caj.seek(current_offset)
                read32 = caj.read(32)
                [
                    image_type_enum,
                    offset_to_image_data,
                    size_of_image_data,
                ] = struct.unpack("iii", read32[0:12])
                if image_type[image_type_enum] != "JPEG":
                    read32 += caj.read(64)
                print(
                    "size of image data = %d (%s)"
                    % (size_of_image_data, image_type[image_type_enum])
                )
                if offset_to_image_data != current_offset + 12:
                    raise SystemExit("unusual image offset")
                print("Page Image Header dump:\n", self.dump(read32), sep="")
                print(
                    "Expected End of Page #%d: 0x%08X"
                    % (i + 1, current_offset + size_of_image_data + 12)
                )
                caj.seek(offset_to_image_data)
                image_data = caj.read(size_of_image_data)
                current_offset = offset_to_image_data + size_of_image_data
                image_name = "image_dump_%04d" % (i + 1)
                if j > 0:
                    image_name = "image_dump_%04d_%04d" % (i + 1, j)
                with open(image_name + ".dat", "wb") as f:
                    f.write(image_data)
                if image_type[image_type_enum] == "JBIG":
                    try:
                        from .dep.jbigdec import SaveJbigAsBmp

                        SaveJbigAsBmp(
                            image_data,
                            size_of_image_data,
                            (image_name + ".bmp").encode("ascii"),
                        )
                    except ImportError:
                        pass
                elif image_type[image_type_enum] == "JBIG2":
                    try:
                        from .dep.jbigdec import SaveJbig2AsBmp

                        SaveJbig2AsBmp(
                            image_data,
                            size_of_image_data,
                            (image_name + ".bmp").encode("ascii"),
                        )
                    except ImportError:
                        pass
                elif image_type[image_type_enum] == "JPEG":
                    with open(image_name + ".jpg", "wb") as f:
                        f.write(image_data)
        print("end 0x%08x" % self._PAGEDATA_OFFSET)

    def dump(self, src, GB=False):
        N = 0
        result = []
        while src:
            s, src = src[:16], src[16:]
            hexa = " ".join(["%02X" % x for x in s])
            gb = ""
            if GB:
                gb += "    "
                for x in range(len(s) >> 1):
                    try:
                        if s[(x << 1) + 1] < 128 and s[(x << 1) + 0] < 128:
                            gb += ".."
                        else:
                            gb += bytes([s[(x << 1) + 1], s[(x << 1) + 0]]).decode(
                                "gbk"
                            )
                    except UnicodeDecodeError:
                        gb += ".."
            s = "".join(printables[x] for x in s)
            result += "%04X   %-*s   %s%s\n" % (N, 16 * 3, hexa, s, gb)
            N += 16
        return "".join(result)

    def _convert_pdf(self, dest):
        copy(self.filename, dest)

    def _convert_kdh(self, dest):
        #  Read KDH file.
        fp = open(self.filename, "rb")
        origin = fp.read()
        fp.close()

        #  Decrypt.
        origin = origin[254:]
        output = []
        keycursor = 0
        for origin_byte in origin:
            output.append(origin_byte ^ KDH_PASSPHRASE[keycursor])
            keycursor += 1
            if keycursor >= len(KDH_PASSPHRASE):
                keycursor = 0
        output = bytes(output)

        #  Remove useless tail data.
        eofpos = output.rfind(b"%%EOF")
        if eofpos < 0:
            raise Exception("%%EOF mark can't be found.")
        output = output[: eofpos + 5]

        #  Write output file.
        fp = open(dest + ".tmp", "wb")
        fp.write(output)
        fp.close()

        # Use mutool to repair xref
        try:
            check_output(["mutool", "clean", dest + ".tmp", dest], stderr=STDOUT)
        except CalledProcessError as e:
            print(e.output.decode("utf-8"))
            raise SystemExit(
                "Command mutool returned non-zero exit status " + str(e.returncode)
            )

        os.remove(dest + ".tmp")



import os
import struct
import sys

import PyPDF2.generic as PDF
from PyPDF2 import PdfFileReader, PdfFileWriter


class Node(object):
    def __init__(self, data, parent=None, lchild=None, rchild=None):
        self.data = data
        self.parent = parent
        self.lchild = lchild
        self.rchild = rchild

    @property
    def level(self):
        return self.data["level"]

    @property
    def index(self):
        return self.data["index"]

    def real_parent(self):
        p = self
        while True:
            c = p
            p = p.parent
            if p.lchild == c:
                return p
            if p.parent is None:
                return None

    def prev(self):
        if self.parent.rchild == self:
            return self.parent
        else:
            return None

    def next(self):
        return self.rchild

    def first(self):
        return self.lchild

    def last(self):
        f = self.first()
        if f is None:
            return None
        r = f
        while r.rchild is not None:
            r = r.rchild
        return r


class BTree(object):
    def __init__(self):
        self.root = Node({"level": 0, "index": 0}, None)
        self.cursor = self.root

    @property
    def current_level(self):
        return self.cursor.level

    def insert_as_lchild(self, node):
        self.cursor.lchild = node
        node.parent = self.cursor
        self.cursor = node

    def insert_as_rchild(self, node):
        self.cursor.rchild = node
        node.parent = self.cursor
        self.cursor = node


def fnd(f, s, start=0):
    fsize = f.seek(0, os.SEEK_END)
    f.seek(0)
    bsize = 4096
    buffer = None
    if start > 0:
        f.seek(start)
    overlap = len(s) - 1
    while True:
        if overlap <= f.tell() < fsize:
            f.seek(f.tell() - overlap)
        buffer = f.read(bsize)
        if buffer:
            pos = buffer.find(s)
            if pos >= 0:
                return f.tell() - (len(buffer) - pos)
        else:
            return -1


def fnd_rvrs(f, s, end=sys.maxsize):
    # find target in reverse direction
    fsize = f.seek(0, os.SEEK_END)
    bsize = 4096
    if len(s) > end:
        raise SystemExit("Too large string size for search.")
    f.seek(fsize - bsize)
    buffer = None
    size = bsize
    if bsize <= end < fsize:
        f.seek(end - bsize)
    elif 0 < end < bsize:
        size = end
        f.seek(0)
    overlap = len(s) - 1
    s = s[::-1]
    while True:
        buffer = f.read(size)
        if buffer:
            buffer = buffer[::-1]
            pos = buffer.find(s)
            if pos >= 0:
                return f.tell() - pos
        if (2 * bsize - overlap) < f.tell():
            f.seek(f.tell() - (2 * bsize - overlap))
            size = bsize
        elif (bsize - overlap) < f.tell():
            size = f.tell() - (bsize - overlap)
            f.seek(0)
        else:
            return -1


def fnd_all(f, s):
    results = []
    last_addr = -len(s)
    while True:
        addr = fnd(f, s, start=last_addr + len(s))
        if addr != -1:
            results.append(addr)
            last_addr = addr
        else:
            return results


def fnd_unuse_no(nos1, nos2):
    unuse_no = -1
    for i in range(99999):
        if (99999 - i not in nos1) and (99999 - i not in nos2):
            unuse_no = 99999 - i
            break
    if unuse_no == -1:
        raise SystemExit("Error on PDF objects numbering.")
    return unuse_no


def make_dest(pdfw, pg):
    d = PDF.ArrayObject()
    try:
        d.append(pdfw.getPage(pg).indirect_ref)
    except AttributeError:
        d.append(pdfw.getPage(pg).indirectRef)
    d.append(PDF.NameObject("/XYZ"))
    d.append(PDF.NullObject())
    d.append(PDF.NullObject())
    d.append(PDF.NullObject())
    return d


def build_outlines_btree(toc):
    tree = BTree()
    for i, t in enumerate(toc):
        t["page"] -= 1  # Page starts at 0.
        t["index"] = i + 1
        node = Node(t)
        if t["level"] > tree.current_level:
            tree.insert_as_lchild(node)
        elif t["level"] == tree.current_level:
            tree.insert_as_rchild(node)
        else:
            while True:
                p = tree.cursor.real_parent()
                tree.cursor = p
                if p.level == t["level"]:
                    tree.insert_as_rchild(node)
                    break
        t["node"] = node


def add_outlines(toc, filename, output):
    build_outlines_btree(toc)
    pdf_out = PdfFileWriter()
    inputFile = open(filename, "rb")
    pdf_in = PdfFileReader(inputFile)
    for p in pdf_in.pages:
        try:
            pdf_out.add_page(p)
        except AttributeError:
            pdf_out.addPage(p)
    toc_num = len(toc)
    if toc_num == 0:  # Just copy if toc empty
        outputFile = open(output, "wb")
        pdf_out.write(outputFile)
        inputFile.close()
        outputFile.close()
        return
    idoix = len(pdf_out._objects) + 1
    idorefs = [PDF.IndirectObject(x + idoix, 0, pdf_out) for x in range(toc_num + 1)]
    ol = PDF.DictionaryObject()
    ol.update(
        {
            PDF.NameObject("/Type"): PDF.NameObject("/Outlines"),
            PDF.NameObject("/First"): idorefs[1],
            PDF.NameObject("/Last"): idorefs[-1],
            PDF.NameObject("/Count"): PDF.NumberObject(toc_num),
        }
    )
    olitems = []
    for t in toc:
        oli = PDF.DictionaryObject()
        oli.update(
            {
                PDF.NameObject("/Title"): PDF.TextStringObject(
                    t["title"].decode("utf-8")
                ),
                PDF.NameObject("/Dest"): make_dest(pdf_out, t["page"]),
            }
        )
        opt_keys = {
            "real_parent": "/Parent",
            "prev": "/Prev",
            "next": "/Next",
            "first": "/First",
            "last": "/Last",
        }
        for k, v in opt_keys.items():
            n = getattr(t["node"], k)()
            if n is not None:
                oli.update({PDF.NameObject(v): idorefs[n.index]})
        olitems.append(oli)
    try:
        pdf_out._add_object(ol)
    except AttributeError:
        pdf_out._addObject(ol)
    for i in olitems:
        try:
            pdf_out._add_object(i)
        except AttributeError:
            pdf_out._addObject(i)
    pdf_out._root_object.update({PDF.NameObject("/Outlines"): idorefs[0]})
    outputFile = open(output, "wb")
    pdf_out.write(outputFile)
    inputFile.close()
    outputFile.close()


# See if the page is N * N images, N images written N times,
# by checking image sizes and within 1 < N <= 10.
# Return True and N if that's the case.
def find_redundant_images(caj, initial_offset, images_per_page):
    sqrts = {
        4: 2,
        9: 3,
        16: 4,
        25: 5,
        36: 6,
        49: 7,
        64: 8,
        81: 9,
        100: 10,
    }

    if not (images_per_page in sqrts.keys()):
        return (False, images_per_page)
    stride = sqrts[images_per_page]
    sizes = []
    current_offset = initial_offset
    for j in range(images_per_page):
        caj.seek(current_offset)
        read32 = caj.read(32)
        [image_type_enum, offset_to_image_data, size_of_image_data] = struct.unpack(
            "iii", read32[0:12]
        )
        if (j >= stride) and (size_of_image_data != sizes[j - stride]):
            return (False, images_per_page)
        sizes.append(size_of_image_data)
        current_offset = offset_to_image_data + size_of_image_data
    # if we reach here, the image sizes seen are [A, B, C ... N, ..., A, B, C ... N] exactly N times.
    return (True, stride)



import os
import pathlib
import sys


def install_context_windows(dry_run: bool):
    binary = sys.argv[0]
    binary_dir = pathlib.Path(binary).parent
    exe = binary_dir / "caj2pdf-ec.exe"

    description = "Convert CAJ to PDF"
    command = f'"{exe}" "%1"'
    if dry_run:
        regedit = f"""\
            Windows Registry Editor Version 5.00

            [HKEY_CLASSES_ROOT\\.caj\\shell\\caj2pdf]
            @="{description}"

            [HKEY_CLASSES_ROOT\\.caj\\shell\\caj2pdf\\command]
            @="{command}"
            """
        print(regedit, file=sys.stderr)
    else:
        import ctypes
        import winreg

        if 1 != ctypes.windll.shell32.IsUserAnAdmin():
            # https://docs.microsoft.com/en-us/windows/win32/api/shellapi/nf-shellapi-shellexecutew
            # https://support.microsoft.com/zh-cn/topic/wd2000-%E5%A6%82%E4%BD%95%E8%B0%83%E7%94%A8-shellexecute-windows-api-%E5%87%BD%E6%95%B0-80da207b-2fa3-ac60-e871-f0a63164bad7
            apifn = ctypes.windll.shell32.ShellExecuteW
            args = (
                # 没有主窗口
                None,
                # 运行 runas 命令
                "runas",
                # 可执行文件
                binary,
                # 运行该程序时的参数
                "install",
                # 工作目录
                os.getcwd(),
                # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-showwindow
                # 激活并显示窗口
                1,
            )
            # TODO 权限获取失败，弹窗消失太快看不清报错。
            print(apifn, args)
            status = apifn(*args)
            if status <= 32:
                raise WindowsError((f"win32api错误，返回 {status}", ("ShellExecuteW", args)))
        else:
            # 如果拥有管理员权限
            reg = winreg.ConnectRegistry(None, winreg.HKEY_CLASSES_ROOT)
            cajshell = winreg.CreateKeyEx(
                key=winreg.HKEY_CLASSES_ROOT,
                sub_key=".caj\\shell\\caj2pdf",
                reserved=0,
                access=winreg.KEY_WRITE,
            )
            winreg.SetValue(
                winreg.HKEY_CLASSES_ROOT,
                ".caj\\shell\\caj2pdf",
                winreg.REG_SZ,
                description,
            )
            cajshellcmd = winreg.CreateKeyEx(
                key=winreg.HKEY_CLASSES_ROOT,
                sub_key=".caj\\shell\\caj2pdf\\command",
                reserved=0,
                access=winreg.KEY_WRITE,
            )
            winreg.SetValue(
                winreg.HKEY_CLASSES_ROOT,
                ".caj\\shell\\caj2pdf\\command",
                winreg.REG_SZ,
                command,
            )
            cajshellcmd.Close()
            cajshell.Close()
            reg.Close()



#!/usr/bin/env python3

#  Copyright 2021 (c) Hin-Tak Leung <htl10@users.sourceforge.net>
#  See The FreeType Project LICENSE for license terms.
#
#  python ctypes module and short program to decode JBIG2 image data in a CAJ file.

#  To build, either libpoppler-based, or libjbig2dec-based (pick only one!):
#
#      cc -Wall `pkg-config --cflags poppler`  -fPIC -shared -o libjbig2codec.so decode_jbig2data.cc   `pkg-config --libs poppler`
#
#      cc -Wall `pkg-config --cflags jbig2dec` -fPIC -shared -o libjbig2codec.so decode_jbig2data_x.cc `pkg-config --libs jbig2dec`
#
#  NOTE(zombie110year,2021/04/20): in this project, just compile them with script file `build.py`

import importlib.resources
import platform
import struct
from ctypes import *

arch = platform.architecture()
if (arch[1] == 'WindowsPE'):
    if (arch[0] == '64bit'):
        with importlib.resources.files(__package__) as pkg_dir:
            libjbig2codec = cdll.LoadLibrary(str(pkg_dir / "bin/libjbig2codec-w64.dll"))
    else:
        with importlib.resources.files(__package__) as pkg_dir:
            libjbig2codec = cdll.LoadLibrary(str(pkg_dir / "bin/libjbig2codec-w32.dll"))
else:
    with importlib.resources.files(__package__) as pkg_dir:
        libjbig2codec = cdll.LoadLibrary(pkg_dir / "bin/libjbig2codec.so")

decode_jbig2data_c    = libjbig2codec.decode_jbig2data_c

decode_jbig2data_c.restype   = c_int
decode_jbig2data_c.argtypes  = [c_void_p, c_int, c_void_p, c_int, c_int, c_int, c_int]

class CImage:
    def __init__(self, buffer):
        self.buffer = buffer
        self.buffer_size=len(buffer)
        (self.width, self.height,
         self.num_planes, self.bits_per_pixel) = struct.unpack("<IIHH", buffer[4:16])
        self.bytes_per_line = ((self.width * self.bits_per_pixel + 31) >> 5) << 2

    def DecodeJbig2(self):
        out = create_string_buffer(self.height * self.bytes_per_line)
        width_in_bytes = (self.width * self.bits_per_pixel + 7) >> 3
        decode_jbig2data_c(self.buffer[48:], self.buffer_size-48, out, self.width, self.height, self.bytes_per_line, width_in_bytes)
        return out

if __name__ == '__main__':
    import os
    import sys

    if len(sys.argv) < 3:
        print("Usage: %s input output" % sys.argv[0])
        sys.exit()

    f = open(sys.argv[1], "rb")
    buffer_size = os.stat(sys.argv[1]).st_size
    buffer = f.read()

    cimage = CImage(buffer)
    out = cimage.DecodeJbig2()

    # PBM is only padded to 8 rather than 32.
    # If the padding is larger, write padded file.
    if (cimage.bytes_per_line > ((cimage.width +7) >> 3)):
        #! bytes_per_line doesn't defined
        cimage.width = cimage.bytes_per_line << 3

    with open(sys.argv[2], "wb") as fout:
        fout.write("P4\n".encode("ascii"))
        fout.write(("%d %d\n" % (cimage.width, cimage.height)).encode("ascii"))
        fout.write(out)
